/**
 * Check is array empty.
 * @param {Array} value Array.
 * @returns Boolean True, if array is empty or not an array type.
 */
Array.isEmpty = function(value) {
  const dataType = Object.getDataType(value);
  if (dataType != "array") {
    return true;
  }

  return (value.length <= 0);
};

/**
 * Create a linear/one-dimension array of objects from an object's property.
 * @param {*} sourceArray Source array.
 * @param {*} propName Property name.
 * @param {*} destArray Destination array.
 */
Array.toFlatten = function (sourceArray, propName, destArray) {
  for (let i=0; i<sourceArray.length; i++) {
    destArray.push(sourceArray[i]);

    if (typeof(sourceArray[i][propName]) != undefined) {
      if (sourceArray[i][propName].length > 0) {
        Array.toFlatten(sourceArray[i][propName], propName, destArray);
      }
    }
  }
};

/**
 * Add elements from an array into this array.
 * @param {Array} array Array.
 * @returns Array.
 */
Array.prototype.addRange = function (array) {
  if (array == null) {
    return this;
  }

  array.forEach(e => {
    this.push(e);
  });

  return this;
};

/**
 * Check all match property value.
 * @param {String} propName Property name.
 * @param {Object} value Property value.
 * @returns Boolean True, if all objects of property name match value.
 */
Array.prototype.all = function (propName, value) {
  for (let i = 0; i < this.length; i++) {
    if (propName == null) {
      if (this[i] != value){
        return false;
      }
    } else {
      if (typeof(this[i][propName]) != "undefined") {
        if (this[i][propName] != value){
          return false;
        }
      }
    }
  }

  return true;
};

/**
 * Check any match property value.
 * @param {String} propName Property name.
 * @param {Object} value Property value.
 * @returns Boolean True, if any objects of property name match value.
 */
Array.prototype.any = function (propName, value) {
  for (let i = 0; i < this.length; i++) {
    if (propName == null) {
      if (this[i] == value){
        return true;
      }
    } else {
      if (typeof(this[i][propName]) != "undefined") {
        if (this[i][propName] == value){
          return true;
        }
      }
    }
  }

  return false;
};

/**
 * Create deep copy of array.
 * @returns Array
 */
Array.prototype.copy = function () {
  return JSON.parse(JSON.stringify(this));
};

/**
 * Count of objects in array with the property value.
 * @param {*} propName Property name.
 * @param {*} value Property value.
 * @returns Integer Number of matches.
 */
Array.prototype.count = function (propName, value) {
  let result = 0;

  for (let i = 0; i < this.length; i++) {
    if (propName == null) {
      if (this[i] == value){
        result++;
      }
    } else {
      if (typeof(this[i][propName]) != "undefined") {
        if (this[i][propName] == value){
          result++;
        }
      }
    }
  }

  return result;
};

/**
 * Count of objects in array with many matches.
 * @param  {Object} filters Array of matches [{propName, value}].
 * @returns Integer Number of matches.
 */
Array.prototype.countMany = function (...filters) {
  let result = 0;

  filters.forEach(e => {
    result += this.count(e.propName, e.value);
  });

  return result;
};

/**
 * Create an empty array of a given length with a given default value.
 * @param {int} length Length of array.
 * @param {object} value Default value.
 * @returns Array.
 */
Array.prototype.create = function (length, value) {
  let result = [];

  for (let i = 0; i < length; i++) {
      result.push(value);
  }

  return result
};

/**
 * Convert this array of objects to a linear/one-dimensional array using a property.
 * @param {*} propName Property name.
 * @returns Array Array.
 */
Array.prototype.flatten = function (propName) {
  let result = [];

  Array.toFlatten(this, propName, result);

  return result;
};

/**
 * Get element at index.
 * @param {Object} index Index of element.
 * @returns Object Element.
 */
Array.prototype.get = function (index) {
  if (index < 0) {
    return "";
  }

  if (index >= this.length) {
    return "";
  }

  return this[index];
};

/**
 * Find index of first occurrence of value in property.
 * @param {string} propName Property name.
 * @param {object} value Property value.
 * @returns Index of element.
 */
Array.prototype.index = function (propName, value) {
  const result = this.indexes(propName, value);
  if (result.length <= 0) {
    return null;
  }

  return result[0];
};

/**
 * Find index of occurrences of value in property in array.
 * @param {string} propName Property name.
 * @param {object} value Property value.
 * @returns Array of indexes.
 */
Array.prototype.indexes = function (propName, value) {
  let result = [];

  for (let i = 0; i < this.length; i++) {
    if (propName == null) {
      if (this[i] == value){
        result.push(i);
      }
    } else {
      if (typeof(this[i][propName]) != "undefined") {
        if (this[i][propName] == value) {
          result.push(i);
        }
      }
    }
  }

  return result;
};

/**
 * Insert an element into an array at index position.
 * @param {int} index Array index.
 * @param {object} item Array element.
 * @returns Array.
 */
Array.prototype.insert = function(index, item) {
  if (index < 0) {
    this.splice(0, 0, item);
  } else if (index >= this.length) {
    this.push(item);
  } else {
    this.splice(index, 0, item);
  }

  return this;
};

/**
 * Concatenates an array if value is not null or whitespace using a separator.
 * @param {string} separator Separator text.
 * @returns string.
 */
Array.prototype.joinIfNotNullOrWhitespace = function (separator) {
  const a = this;

  let result = "";

  for (let i = 0; i < this.length; i++) {
      if (String.isNullOrWhitespace(this[i])) {
          continue;
      }

      if (!String.isNullOrWhitespace(result)) {
          result += separator;
      }

      result += this[i];
  };

  return result;
};

/**
 * Returns first element that matches.
 * @param {string} propName Property name.
 * @param {string} value Property value.
 * @returns Array element.
 */
Array.prototype.first = function (propName, value) {
  for (let i = 0; i < this.length; i++) {
    if (typeof(this[i][propName]) == "undefined") {
      continue;
    }

    if (this[i][propName] == value){
      return this[i];
    }
  }

  return null;
};

/**
 * Traverses an array tree following a property (by name), performing the defined function.
 * @param {string} propName Property name.
 * @param {function} func Function to perform at each element.
 */
Array.prototype.forEachTree = function (propName, func) {
  for (let i=0; i<this.length; i++) {
    if (func(this[i]) === false) {
      return;
    }

    if (this[i][propName].length <= 0) {
      continue;
    }

    if (this[i][propName].forEachTree(propName, func) === false) {
      return;
    }
  }
};

/**
 * Sort an array alphanumerically using the named property (ascending).
 * @param {string} propName Property name.
 * @returns Array.
 */
Array.prototype.orderBy = function (propName) {
  this.sort(function(a, b) {
    if (a[propName] < b[propName]) {
      return -1;
    } else if (a[propName] > b[propName]) {
      return 1;
    } else {
      return 0;
    }
  });

  return this;
};

/**
 * Sort an array alphanumerically using the named property (descending).
 * @param {string} propName Property name.
 * @returns Array.
 */
Array.prototype.orderByDesc = function (propName) {
  this.sort(function(a, b) {
    if (a[propName] < b[propName]) {
      return 1;
    } else if (a[propName] > b[propName]) {
      return -1;
    } else {
      return 0;
    }
  });

  return this;
};

/**
 * Remove an element from an array.
 * @param {object} element Element.
 * @returns Array.
 */
Array.prototype.remove = function (element) {
  let result = [];

  for (let i=0; i<this.length; i++) {
    if (this[i] == element) {
      result.push(i);
    }
  }

  for (let i=(result.length - 1); i>=0; i--) {
    this.removeAt(result[i]);
  }

  return this;
};

/**
 * Remove element at index position.
 * @param {int} index Element index.
 * @returns Array.
 */
Array.prototype.removeAt = function(index) {
  if ((index < 0) || (index >= this.length)) {
    return this;
  }

  this.splice(index, 1);

  return this;
};

/**
 * Remove elements in an array.
 * @param {Array} array Array of elements to be removed.
 * @returns Array.
 */
Array.prototype.removeRange = function (array) {
  for (let i=0; i<array.length; i++) {
    this.remove(array[i]);
  }

  return this;
};

/**
 * Returns array where each element's property does not equal value.
 * @param {string} propName Property name.
 * @param {*} value Property value.
 * @returns Array.
 */
Array.prototype.notSelect = function (propName, value) {
  let result = [];

  for (let i = 0; i < this.length; i++) {
    if (typeof(this[i][propName]) == "undefined") {
      continue;
    }

    if (this[i][propName] != value){
      result.push(this[i]);
    }
  }

  return result;
};

/**
 * Returns array where each element's property equals value.
 * @param {string} propName Property name.
 * @param {*} value Property value.
 * @returns Array.
 */
Array.prototype.select = function (propName, value) {
  let result = [];

  for (let i = 0; i < this.length; i++) {
    if (typeof(this[i][propName]) == "undefined") {
      continue;
    }

    if (this[i][propName] == value){
      result.push(this[i]);
    }
  }

  return result;
};

/**
 * Returns array where each element matches a set of property values.
 * @param  {...any} filters Property-value pairs.
 * @returns Array.
 */
Array.prototype.selectMany = function (...filters) {
  let result = this;

  filters.forEach(e => {
    result = result.select(e.propName, e.value);
  });

  return result;
};

/**
 * Return array of element from array of object by property name.
 * @param {string} propName Property name.
 * @returns Array.
 */
Array.prototype.toList = function (propName) {
  let result = [];

  this.forEach(e => {
    if (typeof(e[propName]) == undefined) {
      return;
    }

    result.push(e[propName]);
  });

  return result;
};

/**
 * Sort this object property using property name.
 * @param {string} childPropName Property name.
 * @param {string} sortPropName Sort by property name.
 */
Array.prototype.sortTree = function (childPropName, sortPropName) {
  this.orderBy(sortPropName);

  for (let i=0; i<this.length; i++) {
    if (this[i][childPropName].length <= 0) {
      continue;
    }

    this[i][childPropName].orderBy(sortPropName);
  }
};